import random
import typing
import pygame
from .agent import AgentState
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources
from .polygon import Polygon
from .util import Point


class Robot(NavigationAgent):
    def __init__(self,
                 start_locations: typing.List[Point.type],
                 open_locations: typing.List[Point.type],
                 navigation: Navigation,
                 view_field: float = 360,
                 max_forward_speed: float = 0.075,
                 max_turning_speed: float = 3.5,
                 ):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=max_forward_speed,
                                 max_turning_speed=max_turning_speed,
                                 view_field=view_field,
                                 size=.05,
                                 sprite_scale=1.9,
                                 polygon_color=(90, 20, 20))
        self.start_locations = start_locations
        self.open_locations = open_locations
        self.last_destination_time = 0

    def reset(self):
        NavigationAgent.reset(self)
        return AgentState(location=random.choice(self.start_locations), direction=180)

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("predator.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_body_polygon() -> Polygon:
        return Polygon([(.02, 0.013), (-.02, 0.013), (-.02, -0.013), (.02, -0.013), (.025, -0.01), (.025, 0.01)])

    def step(self, delta_t: float):
        NavigationAgent.step(self=self,
                             delta_t=delta_t)



from .device import default_device
import shapely as sp
import torch


def normalize_angle(angle):
    return (angle + torch.pi) % (2 * torch.pi) - torch.pi


def polygons_to_sides(polygons):
    vertices = []
    vertices_sides = []

    def add_vertex(vertex):
        i = 0
        for i, v in enumerate(vertices):
            if v.distance(vertex) <= .001:
                break
        else:
            i = len(vertices)
            vertices.append(vertex)
            vertices_sides.append([])
        return i

    sides_vertices = []

    def find_side(sv):
        for i, (a, b) in enumerate(sides_vertices):
            if (a, b) == sv or (b, a) == sv:
                return i
        return -1

    internal_sides = []
    # Process the exterior ring
    for polygon in polygons:
        exterior_coords = list(polygon)
        origin = add_vertex(sp.Point(exterior_coords[0]))
        vertices_sides.append([])
        point_a = origin
        for i in range(1, len(exterior_coords) - 1):
            point_b = add_vertex(sp.Point(exterior_coords[i]))
            i = find_side((point_a, point_b))
            if i == -1:
                side_number = len(sides_vertices)
                sides_vertices.append((point_a, point_b))
                vertices_sides[point_a].append(side_number)
                vertices_sides[point_b].append(side_number)
            else:
                internal_sides.append(i)
            point_a = point_b
        i = find_side((point_a, origin))
        if i == -1:
            side_number = len(sides_vertices)
            sides_vertices.append((point_a, origin))
            vertices_sides[point_a].append(side_number)
            vertices_sides[origin].append(side_number)
        else:
            internal_sides.append(i)

    filtered_sides_vertices = []
    for i, side_vertices in enumerate(sides_vertices):
        if i not in internal_sides:
            filtered_sides_vertices.append(side_vertices)

    sides_centroids = []
    for a, b in filtered_sides_vertices:
        side = sp.LineString([vertices[a], vertices[b]])
        sides_centroids.append(side.centroid)


    vertices = torch.tensor([[vertex.x, vertex.y]
                             for vertex in vertices],
                            device=default_device)

    sides_centroids = torch.tensor([[centroid.x, centroid.y]
                                    for centroid in sides_centroids],
                                   device=default_device)

    filtered_sides_vertices = torch.tensor(filtered_sides_vertices,
                                           device=default_device)

    return vertices, sides_centroids, filtered_sides_vertices


def get_points_angles(src_tensor: torch.tensor,
                      locations: torch.tensor):
    if locations.numel() == 0:
        return torch.tensor([], device=default_device)
    diff = (locations - src_tensor)
    vertices_angles = torch.atan2(diff[:, 1], diff[:, 0])
    return vertices_angles


def get_points_distances(src_tensor: torch.tensor,
                         locations: torch.tensor):
    if locations.numel() == 0:
        return torch.tensor([], device=default_device)
    distances = torch.sum((locations - src_tensor) ** 2, dim=1)
    return distances


def get_intersections(angles: torch.tensor,
                      segments: torch.tensor):

    A = angles.unsqueeze(1)  # Shape (num_rays, 1)
    V1 = segments[:, 0].unsqueeze(0)  # Shape (1, num_segments)
    V2 = segments[:, 1].unsqueeze(0)  # Shape (1, num_segments)
    diff = torch.abs(V2 - V1)

    # Case 1: V1 < V2 and the difference < PI
    case1 = (V1 < V2) & (diff < torch.pi)
    intersect_case1 = case1 & (A >= V1) & (A <= V2)

    # Case 2: V2 < V1 and the difference < PI
    case2 = (V2 < V1) & (diff < torch.pi)
    intersect_case2 = case2 & (A >= V2) & (A <= V1)

    # Case 3: V1 < V2 and the difference > PI
    case3 = (V1 < V2) & (diff > torch.pi)
    intersect_case3 = case3 & ((A >= V2) | (A <= V1))

    # Case 4: V2 < V1 and the difference > PI
    case4 = (V2 < V1) & (diff > torch.pi)
    intersect_case4 = case4 & ((A >= V1) | (A <= V2))

    # Combine the results from all cases
    intersect = intersect_case1 | intersect_case2 | intersect_case3 | intersect_case4

    return intersect



import math
import typing


class CoordinateConverter(object):
    def __init__(self,
                 screen_size: typing.Tuple[int, int],
                 flip_y: bool = False):
        self.screen_size = screen_size
        self.screen_width, self.screen_height = screen_size
        self.flip_y = flip_y
        self.screen_offset = (self.screen_width - self.screen_height) / 2
        self.hexa_ratio = (math.sqrt(3) / 2)

    def scale_to_canonical(self, screen_value: float) -> float:
        return screen_value / self.screen_width

    def scale_from_canonical(self, canonical_value: float) -> float:
        return canonical_value * self.screen_width

    def from_canonical(self, canonical: typing.Union[tuple, float]):
        if isinstance(canonical, float):
            return canonical * self.screen_width

        canonical_x, canonical_y = canonical
        screen_x = canonical_x * self.screen_width
        if self.flip_y:
            screen_y = (1-canonical_y) * self.screen_width - self.screen_offset
        else:
            screen_y = canonical_y * self.screen_width - self.screen_offset
        return screen_x, screen_y

    def to_canonical(self, screen: typing.Union[tuple, float]):
        if isinstance(screen, float):
            return screen / self.screen_width

        screen_x, screen_y = screen
        y = self.screen_height - screen_y + self.screen_offset
        canonical_y = y / self.screen_height * self.hexa_ratio
        canonical_x = screen_x / self.screen_width
        return canonical_x, canonical_y




import typing
from .agent import Agent
from .event import EventDispatcher


class LineOfSight(EventDispatcher):
    def __init__(self):
        EventDispatcher.__init__(self, ["update"])
        self.state: typing.Dict[str, typing.Dict[str, bool]] = {}
        self.agents: typing.Dict[str, Agent] = {}

    def register_agent(self, agent: Agent):
        self.state[agent.name] = {}
        for other_agent_name in self.agents:
            self.register_event(f"{agent.name}_{other_agent_name}")
            self.register_event(f"{other_agent_name}_{agent.name}")
            self.state[agent.name][other_agent_name] = False
            self.state[other_agent_name][agent.name] = False
        self.register_event(f"{agent.name}_update")
        self.agents[agent.name] = agent

    def __setitem__(self,
                    key: tuple,
                    is_visible: bool):

        src_agent, dst_agent = key
        if is_visible:
            self.state[src_agent][dst_agent] = True
            self.__dispatch__(f"{src_agent}_{dst_agent}", self.agents[dst_agent].state)
        else:
            self.state[src_agent][dst_agent] = False

    def update_agent(self,
                     agent_name: str):
        self.__dispatch__(f"{agent_name}_update", self.state[agent_name])

    def update_all(self):
        self.__dispatch__(f"update", self.state)

    def __getitem__(self,
                    key: typing.Union[tuple, str]):
        if isinstance(key, tuple):
            src_agent, dst_agent = key
            return self.state[src_agent][dst_agent]
        else:
            return self.state[key]




import pygame
from .agent import AgentState
from .navigation import Navigation
from .navigation_agent import NavigationAgent
from .resources import Resources
from .polygon import Polygon


class Mouse(NavigationAgent):
    def __init__(self,
                 start_state: AgentState,
                 navigation: Navigation,
                 view_field: float = 360,
                 max_forward_speed: float = 0.5,
                 max_turning_speed: float = 20.0):
        NavigationAgent.__init__(self,
                                 navigation=navigation,
                                 max_forward_speed=max_forward_speed,
                                 max_turning_speed=max_turning_speed,
                                 view_field=view_field,
                                 size=0.04,
                                 sprite_scale=2.0,
                                 polygon_color=(20, 90, 20))
        self.start_state = start_state

    def reset(self):
        NavigationAgent.reset(self)
        return AgentState(location=self.start_state.location, direction=self.start_state.direction)

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("prey.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 270)
        return rotated_sprite

    @staticmethod
    def create_body_polygon() -> Polygon:
        return Polygon([(.015, 0), (0, 0.005), (-.015, 0), (0, -0.005)])




import math
import typing

class Point(object):

    type = typing.Tuple[float, float]

    @staticmethod
    def move(start: "Point.type",
             distance: float,
             direction_degrees: float = None,
             direction_radians: float = None,
             direction_vector: typing.Tuple[float, float] = None) -> typing.Tuple[float, float]:
        if direction_vector is None:
            if direction_radians is None:
                direction_radians = math.radians(direction_degrees)
            start_x, start_y = start
            delta_x = distance * math.cos(direction_radians)
            delta_y = distance * math.sin(direction_radians)
            return start_x + delta_x, start_y + delta_y
        else:
            scaled_vector = Direction.scale_vector(vector=direction_vector,
                                                   scale=distance)
            return Point.add(start=start, vector=scaled_vector)

    @staticmethod
    def add(start: typing.Tuple[float, float],
            vector: typing.Tuple[float, float]) -> typing.Tuple[float, float]:
        return start[0] + vector[0], start[1] + vector[1]

    @staticmethod
    def distance(src: typing.Tuple[float, float],
                 dst: typing.Tuple[float, float]) -> float:
        return math.sqrt((src[0]-dst[0]) ** 2 + (src[1]-dst[1]) ** 2)

    @staticmethod
    def distance2(src: typing.Tuple[float, float],
                  dst: typing.Tuple[float, float]) -> float:
        return (src[0]-dst[0]) ** 2 + (src[1]-dst[1]) ** 2


class Direction:
    vector_type = typing.Tuple[float, float]

    @staticmethod
    def degrees(src: typing.Tuple[float, float],
                dst: typing.Tuple[float, float]) -> float:
        return math.degrees(math.atan2(dst[1] - src[1], dst[0] - src[0]))

    @staticmethod
    def radians(src: typing.Tuple[float, float],
                dst: typing.Tuple[float, float]) -> float:
        return math.atan2(dst[1] - src[1], dst[0] - src[0])

    @staticmethod
    def vector(src: typing.Tuple[float, float],
               dst: typing.Tuple[float, float]) -> typing.Tuple[float, float]:
        if src == dst:
            return 0, 0
        distance = Point.distance(src=src, dst=dst)
        return (dst[0] - src[0]) / distance, (dst[1] - src[1]) / distance

    @staticmethod
    def scale_vector(vector: typing.Tuple[float, float],
                     scale: float = 1):
        return vector[0] * scale, vector[1] * scale

    @staticmethod
    def normalize(direction: float) -> float:
        while direction < -180:
            direction += 360
        while direction > 180:
            direction -= 360
        return direction

    @staticmethod
    def difference(direction1: float,
                   direction2: float) -> float:
        direction1 = Direction.normalize(direction1)
        direction2 = Direction.normalize(direction2)
        difference = direction2 - direction1
        if difference > 180:
            difference -= 360
        if difference < -180:
            difference += 360
        return difference

    @staticmethod
    def error_normalization(direction_error: float):
        pi_err = direction_error / 8
        return 1 / (pi_err * pi_err + 1)


class Line(object):
    type = typing.Tuple[Point.type, Point.type]



import torch

if torch.cuda.is_available():
    default_device = torch.device("cuda")  # Set device to GPU
    print("GPU is available")
else:
    default_device = torch.device("cpu")  # Set device to CPU
    print("GPU is not available, using CPU instead")



import math
import typing
from .util import Point
from .points import Points


class Navigation:
    def __init__(self,
                 locations: typing.List[typing.Optional[Point.type]],
                 paths: typing.List[typing.List[int]],
                 visibility: typing.List[typing.List[typing.List[int]]]):
        self.points = Points(point_list=locations)
        self.paths = paths
        self.visibility = visibility
        self.cache: typing.Dict[typing.Tuple[int, int], typing.List[int]] = {}

    def closest_location(self,
                         location: Point.type) -> int:
        min_dist2 = math.inf
        closest = None
        for i, l in enumerate(self.points.point_list):
            if l is None:
                continue
            dist2 = (l[0] - location[0]) ** 2 + (l[1] - location[1]) ** 2
            if dist2 < min_dist2:
                closest = i
                min_dist2 = dist2
        return closest

    def get_path(self,
                 src: Point.type,
                 dst: Point.type) -> typing.List[Point.type]:
        src_index = self.points.closest(point=src)
        dst_index = self.points.closest(point=dst)
        cache_index = (src_index, dst_index)
        if cache_index in self.cache:
            path_indexes = self.cache[cache_index]
        else:
            current = src_index
            last_step = src_index
            path_indexes = []
            while current is not None and current != dst_index:
                next_step = self.paths[current][dst_index]
                if next_step == current:
                    break
                is_visible = next_step in self.visibility[last_step]
                if not is_visible:
                    path_indexes.append(current)
                    last_step = current
                current = next_step
            path_indexes.append(dst_index)
            self.cache[cache_index] = path_indexes
        return [self.points.point_list[s] for s in path_indexes]



from abc import ABC, abstractmethod
from .coordinate_converter import CoordinateConverter
import typing
import math
from .util import Point


class IPoints(ABC):
    @abstractmethod
    def __init__(self, point_list: typing.List[Point.type]) -> None:
        raise NotImplementedError

    @abstractmethod
    def closest(self, point: Point.type) -> int:
        raise NotImplementedError


class IPolygon(ABC):

    @abstractmethod
    def __init__(self, vertices: typing.List[typing.Tuple[float, float]]):
        raise NotImplementedError

    @abstractmethod
    def contains(self, points):
        raise NotImplementedError

    @abstractmethod
    def intersects(self, other):
        raise NotImplementedError

    @abstractmethod
    def sides(self):
        raise NotImplementedError

    @abstractmethod
    def bounds(self) -> typing.Tuple[float, float, float, float]:
        raise NotImplementedError

    def __iter__(self) -> Point.type:
        side_count = self.sides()
        if side_count:
            for i in range(side_count):
                yield self[i]
            yield self[0]

    @abstractmethod
    def __getitem__(self, item) -> Point.type:
        raise NotImplementedError

    @abstractmethod
    def translate_rotate(self,
                         translation: Point.type,
                         rotation: float,
                         rotation_center: Point.type = (0, 0)) -> "Polygon":
        raise NotImplementedError

    @classmethod
    def regular(cls, center: tuple, diameter: float, angle: float, sides: int):
        radius = diameter / 2
        rotation = math.radians(angle + 90)
        step = math.pi * 2 / sides
        # Generate the points for the hexagon
        points = []
        center_x, center_y = center
        for i in range(sides):
            angle_rad = i * step  # 60 degrees between the points of a hexagon
            x = center_x + radius * math.cos(angle_rad + rotation)
            y = center_y + radius * math.sin(angle_rad + rotation)
            points.append((x, y))

        # Create the regular polygon
        return cls(points)

    def render(self,
               surface,
               coordinate_converter: CoordinateConverter,
               color: typing.Union[typing.Tuple[int, int, int], typing.Tuple[int, int, int, int]]):

        import pygame

        pygame.draw.polygon(surface,
                            color,
                            [coordinate_converter.from_canonical((float(point_x), float(point_y)))
                             for point_x, point_y
                             in self])


class IVisibility(ABC):

    @abstractmethod
    def line_of_sight(self,
                      src: Point.type,
                      dst: Point.type) -> bool:
        raise NotImplementedError

    @abstractmethod
    def line_of_sight_multiple(self,
                               src: Point.type,
                               dst):
        raise NotImplementedError

    @abstractmethod
    def get_visibility_polygon(self,
                               src: Point.type,
                               direction: float,
                               view_field: float = 360) -> IPolygon:
        raise NotImplementedError

    def render(self,
               surface,
               coordinate_converter: CoordinateConverter,
               location: Point.type,
               direction: float,
               view_field: float = 360,
               color: typing.Tuple[int, int, int] = (180, 180, 180)):

        visibility_polygon = self.get_visibility_polygon(src=location,
                                                         direction=direction,
                                                         view_field=view_field)

        visibility_polygon.render(surface=surface,
                                  coordinate_converter=coordinate_converter,
                                  color=color)




import typing
from ..util import Point
from ..interfaces import IPoints
from .device import default_device
import torch


class Points(IPoints):

    def __init__(self, point_list: typing.List[Point.type]) -> None:
        definition = 1000
        x = torch.arange(definition, dtype=torch.float32, device=default_device).repeat(definition, 1)
        y = x.T
        map_points = torch.stack([y, x], dim=2) / definition + 1 / definition / 2
        points_tensor = torch.tensor([point if point else (-1, -1) for point in point_list], device=default_device)
        ext_points = points_tensor.unsqueeze(0).unsqueeze(0).repeat(definition, definition, 1, 1).transpose(0, 2)
        diff = map_points - ext_points
        sdist = diff[:, :, :, 0] ** 2 + diff[:, :, :, 1] ** 2
        self.point_list = point_list
        self.point_map = torch.argmin(sdist, dim=0).cpu().numpy()
        self.definition = definition

    def closest(self, point: Point.type) -> int:
        x, y = point
        index_x = max(0, min(int(x * self.definition), self.definition-1))
        index_y = max(0, min(int(y * self.definition), self.definition-1))
        map_index = self.point_map[index_x, index_y]
        return map_index




import os
from .view import View
from .model import Model
import pygame
import numpy as np


def save_video_output(model: Model,
                      video_folder: str):

    if model.view is None:
        raise ValueError("Model must have a view to save videos.")

    view: View = model.view
    view.gameplay_frames = []

    if not os.path.exists(video_folder):
        os.makedirs(video_folder, exist_ok=True)

    def before_stop(*args):
        video_file = os.path.join(video_folder, f"episode_{model.episode_count:03}.mp4")
        print(f"saving video file {video_file}")
        if view.gameplay_frames:
            from moviepy.editor import ImageSequenceClip
            gameplay_clip = ImageSequenceClip(view.gameplay_frames, fps=int(1/model.time_step))
            gameplay_clip.write_videofile(filename=video_file,
                                          threads=16)
            view.gameplay_frames = []

    frame_count = 0

    def on_frame(surface, _):
        nonlocal frame_count
        frame_count += 1
        frame = np.rot90(pygame.surfarray.array3d(surface))
        frame = np.flipud(frame)
        view.gameplay_frames.append(frame)

    view.add_event_handler("frame", on_frame)
    model.add_event_handler("before_stop", before_stop)



import pygame
import typing
from .resources import Resources
from .util import Point
from .coordinate_converter import CoordinateConverter
from .polygon import Polygon
from .event import EventDispatcher


class AgentState(object):
    def __init__(self, location: Point.type = (0, 0), direction: float = 0):
        self.location = location
        self.direction = direction

    def __iter__(self):
        yield self.location
        yield self.direction

    def update(self,
               distance: float,
               rotation: float) -> "AgentState":
        new_direction = self.direction + rotation
        return AgentState(location=Point.move(start=self.location,
                                              direction_degrees=new_direction,
                                              distance=distance),
                          direction=new_direction)

    def copy(self) -> "AgentState":
        return AgentState(location=self.location,direction=self.direction)


class AgentDynamics(object):
    def __init__(self, forward_speed: float, turn_speed: float):
        self.forward_speed = forward_speed
        self.turn_speed = turn_speed

    def __iter__(self):
        yield self.forward_speed
        yield self.turn_speed

    def change(self, delta_t: float) -> tuple:
        return self.forward_speed * delta_t,  self.turn_speed * delta_t


class Agent(EventDispatcher):
    class RenderMode(object):
        SPRITE = 0
        POLYGON = 1

    def __init__(self,
                 view_field: float = 360,
                 collision: bool = True,
                 size: float = .05,
                 sprite_scale: float = 1.0,
                 polygon_color: typing.Tuple[int, int, int] = (0, 80, 120)):
        self.visible = True
        self.render_mode = Agent.RenderMode.SPRITE
        self.view_field = view_field
        self.state: AgentState = AgentState()
        self.dynamics: AgentDynamics = AgentDynamics(forward_speed=0,
                                                     turn_speed=0)
        self._body_polygon = self.create_body_polygon()
        self.body_polygon = None
        self.visibility_polygon: typing.Optional[Polygon] = None
        self.polygon_color = polygon_color
        self.collision = collision

        self.size = size

        self.sprite = None
        self.sprite_scale = sprite_scale
        EventDispatcher.__init__(self, ["reset", "step"])
        self.on_reset = None
        self.on_step = None
        self.on_start = None
        self.name = ""
        self.model = None
        self.running = False
        self.data = None

    def reset(self) -> AgentState:
        self.__dispatch__("reset")
        self.running = True

    def step(self, delta_t: float) -> None:
        self.__dispatch__("step", delta_t)

    @staticmethod
    def create_sprite() -> pygame.Surface:
        sprite = pygame.image.load(Resources.file("agent.png"))
        rotated_sprite = pygame.transform.rotate(sprite, 90)
        return rotated_sprite

    @staticmethod
    def create_body_polygon() -> Polygon:
        return Polygon.regular((0, 0), .05, 30, sides=6)

    def get_body_polygon(self,
                         state: AgentState = None) -> Polygon:
        # Rotate and then translate the arrow polygon
        if state:
            return self._body_polygon.translate_rotate(translation=state.location, rotation=state.direction)
        else:
            return self._body_polygon.translate_rotate(translation=self.state.location, rotation=self.state.direction)

    def render(self,
               surface: pygame.Surface,
               coordinate_converter: CoordinateConverter):
        if self.visible:
            if self.render_mode == Agent.RenderMode.SPRITE:
                if self.sprite is None:
                    sprite_size = coordinate_converter.scale_from_canonical(self.size) * self.sprite_scale
                    self.sprite = pygame.transform.scale(self.create_sprite(), (sprite_size, sprite_size))
                rotated_sprite = pygame.transform.rotate(self.sprite, self.state.direction)
                width, height = rotated_sprite.get_size()
                screen_x, screen_y = coordinate_converter.from_canonical(self.state.location)
                surface.blit(rotated_sprite, (screen_x - width / 2, screen_y - height / 2))
            else:
                self.body_polygon.render(surface=surface,
                                         coordinate_converter=coordinate_converter,
                                         color=self.polygon_color)



import math

import torch
import typing
from ..interfaces import IPolygon
from ..util import Point
from .device import default_device


class Polygon(IPolygon):
    def __init__(self, vertices):
        if isinstance(vertices, torch.Tensor):
            self.vertices = vertices.to(default_device)
        else:
            self.vertices = torch.tensor(vertices, device=default_device)
        self._sides = len(self.vertices)
        self._vertices_x: torch.tensor = None
        self._vertices_y: torch.tensor = None
        self._edges: torch.tensor = None
        self._bounds = None

    def sides(self):
        return self._sides

    def bounds(self) -> typing.Tuple[float, float, float, float]:
        if self._bounds is None:
            self._bounds = tuple(self.vertices.min(dim=0)[0].tolist()) + tuple(self.vertices.max(dim=0)[0].tolist())
        return self._bounds

    @property
    def vertices_x(self) -> torch.tensor:
        if self._vertices_x is None:
            self._vertices_x = self.vertices[:, 0]
        return self._vertices_x

    @property
    def vertices_y(self) -> torch.tensor:
        if self._vertices_y is None:
            self._vertices_y = self.vertices[:, 1]
        return self._vertices_y

    @property
    def edges(self) -> torch.tensor:
        if self._edges is None:
            self._edges = self.vertices[(torch.arange(self._sides) + 1) % self._sides] - self.vertices
        return self._edges

    def contains(self, points):

        if isinstance(points, Polygon):
            contain_vertices = self.contains(points.vertices)
            return torch.all(contain_vertices)

        if not isinstance(points, torch.Tensor):
            points = torch.tensor(points, device=default_device)

        n_points = points.shape[0]
        n_vertices = self.vertices.shape[0]

        # Extract x and y coordinates of the points
        points_x = points[:, 0]
        points_y = points[:, 1]

        # Repeat polygon coordinates for each point
        poly_x_repeated = self.vertices_x.repeat(n_points, 1)
        poly_y_repeated = self.vertices_y.repeat(n_points, 1)

        # Calculate if edges cross the ray extending to the right of each point
        j = torch.arange(n_vertices) - 1
        vertex_1_y = poly_y_repeated[:, j]
        vertex_2_y = poly_y_repeated
        condition_1 = (vertex_1_y > points_y.unsqueeze(1)) != (vertex_2_y > points_y.unsqueeze(1))
        slope = (poly_x_repeated[:, j] - poly_x_repeated) / (poly_y_repeated[:, j] - poly_y_repeated)
        intercept_x = poly_x_repeated + slope * (points_y.unsqueeze(1) - poly_y_repeated)
        condition_2 = points_x.unsqueeze(1) < intercept_x
        # Determine if the number of crossings is odd or even
        crossings = (condition_1 & condition_2).sum(dim=1)
        inside = crossings % 2 == 1
        return inside

    def intersects(self, other: "Polygon"):
        return self.contains(points=other.vertices).any()

    def __getitem__(self, item) -> typing.Tuple[float, float]:
        return tuple(self.vertices[item, :].tolist())

    def translate_rotate(self,
                         translation: Point.type,
                         rotation: float,
                         rotation_center: Point.type = (0, 0)) -> "Polygon":

        t = torch.tensor(translation, device=default_device)
        rc = torch.tensor(rotation_center, device=default_device)
        r = torch.tensor(math.radians(-rotation), device=default_device)
        vertices = self.vertices.clone() - rc
        rotation_matrix = torch.tensor([
            [torch.cos(r), -torch.sin(r)],
            [torch.sin(r), torch.cos(r)]
        ], dtype=torch.float32, device=default_device)
        rotated_points = torch.matmul(vertices, rotation_matrix)
        rotated_points += rc + t
        return Polygon(vertices=rotated_points)

    def area(self) -> float:
        # Extract x and y coordinates
        x = self.vertices[:, 0]
        y = self.vertices[:, 1]
        # Apply the Shoelace formula
        area = 0.5 * torch.abs(torch.sum(x[:-1] * y[1:]) + x[-1] * y[0] - torch.sum(y[:-1] * x[1:]) - y[-1] * x[0])
        return float(area)
    


import time
import typing
import shapely as sp

from .util import Point
from .agent import Agent, AgentState
from .visibility import Visibility
from .polygon import Polygon
from .event import EventDispatcher
from .line_of_sight import LineOfSight


class Model(EventDispatcher):

    def __init__(self,
                 world_name: str,
                 arena: Polygon,
                 occlusions: typing.List[Polygon],
                 time_step: float = 0.1,
                 real_time: bool = False,
                 render: bool = False,
                 agent_point_of_view: str = "",
                 agent_render_mode: Agent.RenderMode = Agent.RenderMode.SPRITE,
                 max_line_of_sight_distance: float = 1.0):
        self.world_name = world_name
        self.arena = arena
        self.occlusions = occlusions
        self.time_step = time_step
        self.real_time = real_time
        self.render = render
        self.agent_point_of_view = agent_point_of_view
        self.agent_render_mode = agent_render_mode
        self.max_line_of_sight_distance = max_line_of_sight_distance
        self.agents: typing.Dict[str, Agent] = {}
        self.visibility = Visibility(arena=self.arena, occlusions=self.occlusions)
        self.last_step = None
        self.time: float = 0
        self.running = False
        self.episode_count = 0
        self.step_count = 0
        self.view: typing.Optional["View"] = None
        self.paused = False
        self.line_of_sight = LineOfSight()
        EventDispatcher.__init__(self, ["before_step",
                                        "after_step",
                                        "before_stop",
                                        "after_stop",
                                        "before_reset",
                                        "after_reset",
                                        "agents_states_update",
                                        "close",
                                        "pause"])
        if self.render:
            self.occlusion_color = (50, 50, 50)
            self.arena_color = (210, 210, 210)
            self.visibility_color = (255, 255, 255)
            from .view import View
            self.view = View(model=self)
            self.view.add_event_handler("quit", self.close)

            def render_occlusions(surface, coordinate_converter):
                for occlusion in self.occlusions:
                    occlusion.render(surface=surface,
                                     coordinate_converter=coordinate_converter,
                                     color=self.occlusion_color)

            def render_arena(surface, coordinate_converter):
                self.arena.render(surface=surface,
                                  coordinate_converter=coordinate_converter,
                                  color=self.arena_color)

            self.view.add_render_step(render_step=render_arena, z_index=0)
            if agent_point_of_view == "":
                self.view.add_render_step(render_step=render_occlusions, z_index=30)
            else:
                self.view.add_render_step(render_step=render_occlusions, z_index=1030)

            self.render_agent_visibility = agent_point_of_view

            def render_visibility(surface, coordinate_converter):
                if self.render_agent_visibility == "":
                    return
                visibility_polygon = self.agents[self.render_agent_visibility].visibility_polygon
                visibility_polygon.render(surface=surface,
                                          coordinate_converter=coordinate_converter,
                                          color=self.visibility_color)

            def render_hidden_area(surface, coordinate_converter):
                import pygame
                if self.render_agent_visibility == "":
                    return
                visibility_polygon = self.agents[self.render_agent_visibility].visibility_polygon
                mask_surface = pygame.Surface(surface.get_size(), pygame.SRCALPHA)
                mask_surface.fill((0, 0, 0, 255))
                self.arena.render(surface=mask_surface,
                                  coordinate_converter=coordinate_converter,
                                  color=self.arena_color + (255,))
                visibility_polygon.render(surface=mask_surface,
                                          coordinate_converter=coordinate_converter,
                                          color=(0, 0, 0, 0))
                surface.blit(mask_surface, (0, 0))

            def key_down(key):
                import pygame
                if key == pygame.K_r:
                    if self.agent_render_mode == Agent.RenderMode.SPRITE:
                        self.agent_render_mode = Agent.RenderMode.POLYGON
                    else:
                        self.agent_render_mode = Agent.RenderMode.SPRITE
                    for agent_name, agent in self.agents.items():
                        agent.render_mode = self.agent_render_mode
                elif key == pygame.K_v:
                    agent_names = list(self.agents.keys())
                    if self.render_agent_visibility == "":
                        self.render_agent_visibility = agent_names[0]
                    else:
                        current_agent = agent_names.index(self.render_agent_visibility)
                        if current_agent == len(agent_names) - 1:
                            self.render_agent_visibility = ""
                        else:
                            self.render_agent_visibility = agent_names[current_agent + 1]
                elif key == pygame.K_p:
                    self.pause()
                elif key == pygame.K_n:
                    for agent_name, agent in self.agents.items():
                        if hasattr(agent, "render_path"):
                            agent.render_path = not agent.render_path

            self.view.add_render_step(render_visibility, z_index=20)
            if agent_point_of_view:
                self.view.add_render_step(render_hidden_area, z_index=1000)

            self.view.add_event_handler("key_down", key_down)

    def get_agents_state(self) -> typing.Dict[str, AgentState]:
        agents_state: typing.Dict[str, AgentState] = {}
        for agent_name, agent in self.agents.items():
            agents_state[agent_name] = agent.state.copy()
        return agents_state

    def set_agents_state(self,
                         agents_state: typing.Dict[str, AgentState],
                         agents_body_polygons: typing.Dict[str, Polygon] = None,
                         delta_t: float = 0):
        for agent_name, agent_state in agents_state.items():
            agent = self.agents[agent_name]
            agent.state = agent_state
            if agents_body_polygons:
                agent.body_polygon = agents_body_polygons[agent_name]
            else:
                agent.body_polygon = agent.get_body_polygon(state=agent_state)

            agent.visibility_polygon = self.visibility.get_visibility_polygon(src=agent_state.location,
                                                                              direction=agent_state.direction,
                                                                              view_field=agent.view_field)

            self.__dispatch__(f"agent_{agent_name}_state_update", agent_state)

        for agent_name, agent in self.agents.items():
            agent_visibility_polygon = agent.visibility_polygon
            for other_agent_name, other_agent in self.agents.items():
                if other_agent_name != agent_name:
                    if Point.distance(agent.state.location, other_agent.state.location) <= self.max_line_of_sight_distance:
                        has_line_of_sight = agent_visibility_polygon.intersects(other_agent.body_polygon)
                    else:
                        has_line_of_sight = False
                    self.line_of_sight[agent_name, other_agent_name] = has_line_of_sight
        self.__dispatch__("agents_states_update", agents_state, self.line_of_sight)

    def pause(self):
        self.paused = not self.paused
        self.__dispatch__("pause", self)

    def add_agent(self, name: str, agent: Agent):
        agent.name = name
        agent.model = self
        agent.render_mode = self.agent_render_mode
        self.register_event(f"agent_{name}_state_update")
        self.line_of_sight.register_agent(agent=agent)
        self.agents[name] = agent
        if self.render:
            self.view.add_render_step(agent.render, z_index=100)

    def reset(self,
              agents_state: typing.Dict[str, AgentState] = None):
        if self.running:
            self.stop()
        self.__dispatch__("before_reset", self)
        self.running = True
        self.episode_count += 1
        agents_start_state: typing.Dict[str, AgentState] = {}
        agents_body_polygon: typing.Dict[str, Polygon] = {}
        for name, agent in self.agents.items():
            agent_reset_state = agent.reset()
            agent_state = agents_state[name] if agents_state and name in agents_state else agent_reset_state
            agents_start_state[name] = agent_state
            agents_body_polygon[name] = agent.get_body_polygon(state=agent_state)

        self.set_agents_state(agents_state=agents_start_state)
        self.last_step = time.time()
        self.step_count = 0
        self.__dispatch__("after_reset", self)

    def stop(self):
        if not self.running:
            return
        self.__dispatch__("before_stop", self)
        self.running = False
        self.__dispatch__("after_stop", self)

    def is_valid_state(self, agent_polygon: sp.Polygon, collisions: bool) -> bool:
        if not self.arena.contains(agent_polygon):
            return False
        if collisions:
            for occlusion in self.occlusions:
                if agent_polygon.intersects(occlusion):
                    return False
        return True

    def step(self) -> float:
        if not self.running:
            return 0

        if self.paused:
            return 0

        self.__dispatch__("before_step", self)

        if self.real_time:
            while self.last_step + self.time_step > time.time():
                pass

        self.last_step = time.time()
        new_states: typing.Dict[str, AgentState] = {}
        new_body_polygons: typing.Dict[str, Polygon] = {}
        for name, agent in self.agents.items():
            dynamics = agent.dynamics
            distance, rotation = dynamics.change(delta_t=self.time_step)
            new_state = agent.state.update(rotation=rotation,
                                           distance=distance)
            agent_polygon = agent.get_body_polygon(state=new_state)
            if not self.is_valid_state(agent_polygon=agent_polygon,
                                       collisions=agent.collision): #try only rotation
                new_state = agent.state.update(rotation=rotation,
                                               distance=0)
                agent_polygon = agent.get_body_polygon(state=new_state)
                if not self.is_valid_state(agent_polygon=agent_polygon,
                                           collisions=agent.collision): #try only translation
                    new_state = agent.state.update(rotation=0,
                                                   distance=distance)
                    agent_polygon = agent.get_body_polygon(state=new_state)
                    if not self.is_valid_state(agent_polygon=agent_polygon,
                                               collisions=agent.collision):
                        new_state = agent.state
                        agent_polygon = agent.body_polygon
            new_states[name] = new_state
            new_body_polygons[name] = agent_polygon
        self.set_agents_state(agents_state=new_states, agents_body_polygons=new_body_polygons)
        for name, agent in self.agents.items():
            agent.step(delta_t=self.time_step)
        self.time += self.time_step
        self.step_count += 1
        self.__dispatch__("after_step", self)
        return self.time_step

    def close(self):
        if self.running:
            self.stop()
        self.__dispatch__("close", self)

    def __del__(self):
        self.close()



import shapely as sp
import math
from ..util import Point
from ..util import Direction

def theta_in_between(theta: float, start: float, end: float) -> bool:
    if end > start:
        return start < theta < end
    return theta > start or theta < end


def distance2(point1: sp.Point, point2: sp.Point):
    return Point.distance2(src=(point1.x, point1.y), dst=(point2.x, point2.y))


def atan2(src: sp.Point,
          dst: sp.Point):
    return math.atan2(dst.y - src.y, dst.x - src.x)


def move(src: sp.Point,
         theta: float,
         dist: float):
    return sp.Point(src.x + math.cos(theta) * dist, src.y + math.sin(theta) * dist)


def polygons_to_sides(polygons):
    """
    Break a Polygon into individual LineStrings representing its edges.

    Parameters:
    polygon (Polygon): The input Polygon object.
    include_holes (bool): If True, include the interior boundaries (holes) as well.

    Returns:
    List[LineString]: A list of LineString objects representing the edges of the polygon.
    """
    vertices = []
    vertices_sides = []

    def add_vertex(vertex):
        i = 0
        for i, v in enumerate(vertices):
            if v.distance(vertex) <= .001:
                break
        else:
            i = len(vertices)
            vertices.append(vertex)
            vertices_sides.append([])
        return i

    sides_vertices = []

    def find_side(sv):
        for i, (a, b) in enumerate(sides_vertices):
            if (a, b) == sv or (b, a) == sv:
                return i
        return -1

    internal_sides = []
    # Process the exterior ring
    for polygon in polygons:
        exterior_coords = list(polygon)
        origin = add_vertex(sp.Point(exterior_coords[0]))
        vertices_sides.append([])
        point_a = origin
        for i in range(1, len(exterior_coords) - 1):
            point_b = add_vertex(sp.Point(exterior_coords[i]))
            i = find_side((point_a, point_b))
            if i == -1:
                side_number = len(sides_vertices)
                sides_vertices.append((point_a, point_b))
                vertices_sides[point_a].append(side_number)
                vertices_sides[point_b].append(side_number)
            else:
                internal_sides.append(i)
            point_a = point_b
        i = find_side((point_a, origin))
        if i == -1:
            side_number = len(sides_vertices)
            sides_vertices.append((point_a, origin))
            vertices_sides[point_a].append(side_number)
            vertices_sides[origin].append(side_number)
        else:
            internal_sides.append(i)

    filtered_sides_vertices = []
    for i, side_vertices in enumerate(sides_vertices):
        if i not in internal_sides:
            filtered_sides_vertices.append(side_vertices)

    sides = []
    sides_centroids = []
    for a, b in filtered_sides_vertices:
        side = sp.LineString([vertices[a], vertices[b]])
        sides.append(side)
        sides_centroids.append(side.centroid)

    return sides, sides_centroids, vertices, filtered_sides_vertices



import random
import typing
from ..util import Point
from ..model import Model
from ..agent import AgentState, CoordinateConverter
from ..mouse import Mouse
from ..robot import Robot
from ..cellworld_loader import CellWorldLoader
from ..agent import Agent
import enum


class Oasis(Model):

    class PointOfView(enum.Enum):
        TOP = ""
        PREY = "prey"
        PREDATOR = "predator"

    def __init__(self,
                 world_name: str,
                 goal_locations: typing.List[typing.Tuple[float, float]],
                 goal_sequence_generator: typing.Callable[[None], typing.List[int]] = None,
                 use_predator: bool = True,
                 puff_cool_down_time: float = .5,
                 goal_time: float = 1,
                 puff_threshold: float = .1,
                 goal_threshold: float = .1,
                 time_step: float = .025,
                 real_time: bool = False,
                 render: bool = False,
                 point_of_view: PointOfView = PointOfView.TOP,
                 agent_render_mode: Agent.RenderMode = Agent.RenderMode.SPRITE,
                 max_line_of_sight_distance: float = 1.0):

        if goal_sequence_generator is None:
            goal_sequence_generator = lambda: random.sample(range(0, len(goal_locations)), 3)

        self.start_location = (.05, .5)
        self.goal_locations = goal_locations
        self.goal_time = goal_time
        self.goal_sequence_generator = goal_sequence_generator
        self.use_predator = use_predator
        self.puff_cool_down_time = puff_cool_down_time
        self.puff_threshold = puff_threshold
        self.goal_threshold = goal_threshold
        self.loader = CellWorldLoader(world_name=world_name)
        self.goal_location = None
        self.goal_sequence: typing.List[int] = []

        Model.__init__(self,
                       world_name=world_name,
                       arena=self.loader.arena,
                       occlusions=self.loader.occlusions,
                       time_step=time_step,
                       real_time=real_time,
                       render=render,
                       agent_render_mode=agent_render_mode,
                       agent_point_of_view=point_of_view.value,
                       max_line_of_sight_distance=max_line_of_sight_distance)

        if use_predator:
            self.predator = Robot(start_locations=self.loader.robot_start_locations,
                                  open_locations=self.loader.open_locations,
                                  navigation=self.loader.navigation)

            self.add_agent("predator", self.predator)

        self.prey = Mouse(start_state=AgentState(location=self.start_location,
                                                 direction=0),
                          navigation=self.loader.navigation)

        self.add_agent("prey", self.prey)

        if self.render:
            import pygame

            def render_active_goal(surface: pygame.Surface,
                                   coordinate_converter: CoordinateConverter):
                if self.goal_location is None:
                    return
                for goal_location in self.goal_locations:
                    screen_goal_location = coordinate_converter.from_canonical(goal_location)
                    radius = coordinate_converter.from_canonical(.0125)
                    goal_color = (0, 255, 0) if self.goal_location == goal_location else (255, 0, 0)
                    pygame.draw.circle(surface,
                                       color=goal_color,
                                       center=screen_goal_location,
                                       radius=radius)

            self.view.add_render_step(render_active_goal)

            if use_predator:
                def render_puff_area(surface: pygame.Surface,
                                     coordinate_converter: CoordinateConverter):
                    predator_location = coordinate_converter.from_canonical(self.predator.state.location)
                    puff_area_size = self.puff_threshold * coordinate_converter.screen_width
                    puff_location = predator_location[0] - puff_area_size, predator_location[1] - puff_area_size
                    puff_area_surface = pygame.Surface((puff_area_size * 2, puff_area_size * 2), pygame.SRCALPHA)
                    puff_area_color = (255, 0, 0, 60) if self.puff_cool_down > 0 else (0, 0, 255, 60)
                    pygame.draw.circle(puff_area_surface,
                                       color=puff_area_color,
                                       center=(puff_area_size, puff_area_size),
                                       radius=puff_area_size)
                    surface.blit(puff_area_surface,
                                 puff_location)
                    pygame.draw.circle(surface=surface,
                                       color=(0, 0, 255),
                                       center=predator_location,
                                       radius=puff_area_size,
                                       width=2)

                self.view.add_render_step(render_puff_area, z_index=90)

        self.puffed: bool = False
        self.puff_cool_down: float = 0
        self.goal_achieved: bool = True
        self.goal_achieved_time: float = 0
        self.predator_prey_distance: float = 1
        self.prey_goal_distance: float = 0
        self.puff_count = 0

    def __update_state__(self,
                         delta_t: float = 0):
        if self.use_predator and self.puff_cool_down <= 0:
            self.predator_prey_distance = Point.distance(self.prey.state.location,
                                                         self.predator.state.location)
            if self.visibility.line_of_sight(self.prey.state.location, self.predator.state.location):
                if self.predator_prey_distance <= self.puff_threshold:
                    self.puffed = True
                    self.puff_count += 1
                    self.puff_cool_down = self.puff_cool_down_time

                self.predator.set_destination(self.prey.state.location)

            if not self.predator.path:
                self.predator.set_destination(random.choice(self.loader.open_locations))

        if delta_t < self.puff_cool_down:
            self.puff_cool_down -= delta_t
        else:
            self.puff_cool_down = 0

        self.prey_goal_distance = Point.distance(self.goal_location, self.prey.state.location)

        if self.prey_goal_distance <= self.goal_threshold:
            self.goal_achieved_time += delta_t
            if self.goal_achieved_time >= self.goal_time:
                self.__update_goal__()
                self.goal_achieved_time = 0
                self.goal_achieved = False
                if self.goal_location is None:
                    self.stop()
            else:
                self.goal_achieved = True

    def __on_quit__(self):
        self.stop()

    def __update_goal__(self):
        if self.goal_sequence:
            self.goal_location = self.goal_locations[self.goal_sequence.pop(0)]
        else:
            if self.goal_location == self.start_location:
                self.goal_location = None
            else:
                self.goal_location = self.start_location

    def reset(self):
        Model.reset(self)
        self.goal_achieved = False
        self.puff_count = 0
        self.goal_sequence = self.goal_sequence_generator()
        self.__update_goal__()
        self.__update_state__()

    def step(self) -> float:
        delta_t = Model.step(self)
        if self.render:
            self.view.render()
        self.__update_state__(delta_t=delta_t)
        return delta_t


from .view import View
from .model import Model
from .agent import Agent, AgentDynamics, AgentState, CoordinateConverter
from .resources import Resources
from .robot import Robot
from .mouse import Mouse
from .navigation import Navigation
from .util import Point, Direction, Line
from .cellworld_loader import CellWorldLoader
from .visibility import Visibility
from .tasks import BotEvade, Oasis, DualEvade
from .video import save_video_output
from .polygon import Polygon
from .points import Points
from .log import save_log_output
from .event import EventDispatcher
from .line_of_sight import LineOfSight



import torch
import math
import typing
import shapely as sp
from .geometry import polygons_to_sides, normalize_angle, get_points_angles, get_points_distances, get_intersections
from .device import default_device
from .polygon import Polygon
from ..interfaces import IVisibility
from ..util import Point


class Visibility(IVisibility):
    def __init__(self, arena: sp.Polygon, occlusions: typing.List[sp.Polygon]):
        self.occlusion_vertices, self.occlusions_centroids, self.occlusions_vertices_indices = polygons_to_sides(occlusions)
        self.wall_vertices, self.walls_centroids, self.walls_vertices_indices = polygons_to_sides([arena])

    def line_of_sight_multiple(self,
                               src: Point.type,
                               dst):
        polygon = self.get_visibility_polygon(src=src, direction=0)
        return polygon.contains(dst)

    def line_of_sight(self,
                      src: Point.type,
                      dst: Point.type) -> bool:

        src_tensor = torch.tensor(src, device=default_device)
        dst_tensor = torch.tensor(dst, device=default_device)

        occlusions_vertices_angles = get_points_angles(src_tensor=src_tensor,
                                                       locations=self.occlusion_vertices)

        occlusions_segments_angles = self.__get_segments_vertices_angles__(vertices_angles=occlusions_vertices_angles,
                                                                           segments=self.occlusions_vertices_indices)

        ray = get_points_angles(src_tensor=src_tensor, locations=dst_tensor.unsqueeze(0))

        intersections = get_intersections(angles=ray,
                                          segments=occlusions_segments_angles)[0, :]

        if not intersections.any():
            return True

        occlusions_distances = get_points_distances(src_tensor=src_tensor,
                                                    locations=self.occlusions_centroids[intersections])

        dst_distance = get_points_distances(src_tensor=src_tensor,
                                            locations=dst_tensor.unsqueeze(0))

        if (dst_distance < occlusions_distances).all():
            return True
        else:
            return False

    def __get_segments_vertices_angles__(self,
                                         vertices_angles: torch.tensor,
                                         segments: torch.tensor):
        if segments.numel() == 0:
            return torch.tensor([], device=default_device)
        return torch.stack([normalize_angle((vertices_angles[segments[:, 0]])),
                            normalize_angle((vertices_angles[segments[:, 1]]))], dim=1).to(default_device)

    def __get_segments_vertices_points__(self,
                                         vertices_points: torch.tensor,
                                         segments: torch.tensor):
        if segments.numel() == 0:
            return torch.tensor([], device=default_device)
        return torch.stack([vertices_points[segments[:, 0]],
                            vertices_points[segments[:, 1]]], dim=1).to(default_device)

    def get_visibility_polygon(self,
                               src: Point.type,
                               direction: float,
                               view_field: float = 360):

        theta = math.radians((direction + 180) % 360) - math.pi

        src_tensor = torch.tensor(src, device=default_device)

        occlusions_vertices_angles = get_points_angles(src_tensor=src_tensor,
                                                       locations=self.occlusion_vertices)

        occlusions_segments_angles = self.__get_segments_vertices_angles__(vertices_angles=occlusions_vertices_angles,
                                                                           segments=self.occlusions_vertices_indices)

        occlusions_segments_points = self.__get_segments_vertices_points__(vertices_points=self.occlusion_vertices,
                                                                           segments=self.occlusions_vertices_indices)

        occlusions_distances = get_points_distances(src_tensor=src_tensor,
                                                    locations=self.occlusions_centroids)

        walls_vertices_angles = get_points_angles(src_tensor=src_tensor,
                                                  locations=self.wall_vertices)

        walls_segments_angles = self.__get_segments_vertices_angles__(vertices_angles=walls_vertices_angles,
                                                                      segments=self.walls_vertices_indices)

        walls_segments_points = self.__get_segments_vertices_points__(vertices_points=self.wall_vertices,
                                                                      segments=self.walls_vertices_indices)

        walls_distances = get_points_distances(src_tensor=src_tensor,
                                               locations=self.walls_centroids) + 1.0 # this ensures walls are checked last for intersections

        vertices_angles = torch.cat((occlusions_vertices_angles, walls_vertices_angles), dim=0)
        segments_vertices_angles = torch.cat((occlusions_segments_angles, walls_segments_angles), dim=0)
        segments_vertices_points = torch.cat((occlusions_segments_points, walls_segments_points), dim=0)
        segments_distances = torch.cat((occlusions_distances, walls_distances), dim=0)
        rays = torch.cat((vertices_angles - .001, vertices_angles, vertices_angles + .001), dim=0)

        if view_field < 360:
            d = math.radians(view_field) / 2
            lb = theta - d
            ub = theta + d
            if lb < -math.pi:
                lb += 2 * math.pi
                filtered_thetas = rays[torch.logical_or(rays > lb, rays < ub)]
            elif ub > math.pi:
                ub -= 2 * math.pi
                filtered_thetas = rays[torch.logical_or(rays > lb, rays < ub)]
            else:
                filtered_thetas = rays[(rays > lb) & (rays < ub)]

            rays = torch.cat((torch.tensor([lb, ub], device=default_device), filtered_thetas), dim=0)

        ordered_rays, rays_indices = torch.sort(rays)

        if view_field < 360:
            # saves_the_index of the lower and upper bound ray indices
            lower_bound_index = torch.nonzero((rays_indices == 0), as_tuple=True)[0]
            upper_bound_index = torch.nonzero((rays_indices == 1), as_tuple=True)[0]

        intersections = get_intersections(angles=ordered_rays,
                                          segments=segments_vertices_angles)

        segments_distances = segments_distances.expand_as(intersections)

        intersected_walls_distances = torch.where(intersections,
                                                  segments_distances,
                                                  math.inf).to(default_device)

        closest_intersected_walls = torch.argmin(intersected_walls_distances, dim=1)

        # filters unnecessary vertices
        non_colinear = torch.zeros_like(closest_intersected_walls, dtype=torch.bool)
        # Set the first and last elements to True
        non_colinear[0] = True
        non_colinear[-1] = True
        # Compare each element with its previous and next neighbors
        non_colinear[1:-1] = (closest_intersected_walls[1:-1] != closest_intersected_walls[:-2]) | (closest_intersected_walls[1:-1] != closest_intersected_walls[2:])

        if view_field < 360:
            # do not filter the bounds rays
            non_colinear[lower_bound_index] = True
            non_colinear[upper_bound_index] = True

        filtered_intersected_walls = closest_intersected_walls[non_colinear]
        filtered_rays = ordered_rays[non_colinear]
        filtered_rays_indices = rays_indices[non_colinear]

        xy_v1 = segments_vertices_points[filtered_intersected_walls, 0, :]
        xy_v2 = segments_vertices_points[filtered_intersected_walls, 1, :]

        direction_x = torch.cos(filtered_rays)
        direction_y = torch.sin(filtered_rays)

        Q0_x, Q0_y, Q1_x, Q1_y = xy_v1[:, 0], xy_v1[:, 1], xy_v2[:, 0], xy_v2[:, 1]

        segment_direction_x = Q1_x - Q0_x
        segment_direction_y = Q1_y - Q0_y

        src_tensor_x, src_tensor_y = src_tensor
        A = torch.stack([direction_x, -segment_direction_x, direction_y, -segment_direction_y], dim=1).reshape(-1, 2, 2)
        b = torch.stack([Q0_x - src_tensor_x, Q0_y - src_tensor_y], dim=1).unsqueeze(2)

        solution = torch.linalg.solve(A, b)
        t = solution[:, 0, 0]

        # Compute the intersection points
        intersection_x = src_tensor_x + t * direction_x
        intersection_y = src_tensor_y + t * direction_y

        # Combine the angle and intersection points into a single tensor

        result = torch.stack([intersection_x, intersection_y], dim=1).to(default_device)

        if view_field < 360:
            # adds a vertex on the source location
            index = torch.nonzero((filtered_rays_indices == 0), as_tuple=True)[0]
            result = torch.cat([result[:index, :], src_tensor.unsqueeze(0), result[index:, :]], dim=0)
        return Polygon(result)




import typing


class EventDispatcher:
    def __init__(self, events: typing.List[str]):
        self.event_handlers: typing.Dict[str, typing.List[typing.Callable]] = {event_name: [] for event_name in events}

    def register_event(self, event_name: str):
        self.event_handlers[event_name] = []

    def __dispatch__(self, event_name: str, *args):
        for event_handler in self.event_handlers[event_name]:
            event_handler(*args)

    def add_event_handler(self, event_name: str, handler: typing.Callable):
        if event_name not in self.event_handlers:
            raise f"Event {event_name} not registered"
        self.event_handlers[event_name].append(handler)




import typing
from ..util import Point
from ..interfaces import IPoints
import numpy as np


class Points(IPoints):

    def __init__(self, point_list: typing.List[Point.type]) -> None:
        self.point_list = point_list
        self.points_array = np.array([point if point else (-1, -1) for point in point_list])

    def closest(self, point: Point.type) -> Point.type:
        point_array = np.array(point)
        distances = np.sum((self.points_array - point_array) ** 2, axis=1)
        return snp.argmin(distances)



import os.path

import cellworld as cw
from .model import Model


def save_log_output(model: Model,
                    experiment_name: str,
                    log_folder: str,
                    save_checkpoint: bool = True):

    if not os.path.exists(log_folder):
        os.makedirs(log_folder, exist_ok=True)

    experiment = cw.Experiment(name=experiment_name,
                               world_configuration_name="hexagonal",
                               world_implementation_name="mice",
                               duration=0,
                               occlusions=model.world_name)

    frame = 0

    def after_reset(*args):
        nonlocal frame
        experiment.episodes.append(cw.Episode())
        frame = 0

    def after_stop(*args):
        import os
        output_file = os.path.join(log_folder, f"{experiment_name}.json")
        print(f"saving log file {output_file}")
        experiment.save(output_file)

    def after_step(*args):
        nonlocal frame
        episode: cw.Episode = experiment.episodes[-1]
        for agent_name, agent in model.agents.items():
            agent_step = cw.Step(time_stamp=model.time,
                                 location=cw.Location(*agent.state.location),
                                 rotation=90-agent.state.direction,
                                 agent_name=agent_name,
                                 frame=frame)
            episode.trajectories.append(agent_step)
        frame += 1

    if "puff" in model.event_handlers:
        def puff(_):
            episode: cw.Episode = experiment.episodes[-1]
            episode.captures.append(model.step_count)

        model.add_event_handler("puff", puff)

    model.add_event_handler("after_reset", after_reset)
    model.add_event_handler("after_step", after_step)
    if save_checkpoint:
        model.add_event_handler("after_stop", after_stop)
    else:
        model.add_event_handler("close", after_stop)



import typing
from ..interfaces import IPolygon
import shapely.affinity as spa
import shapely as sp
from ..util import Point


class Polygon(IPolygon):

    def __init__(self, vertices):
        self._sides = len(vertices)
        self._vertices = None
        self._points = None
        self._polygon = None
        self._bounds = None
        if vertices:
            if isinstance(vertices[0], sp.Point):
                self._points = vertices
            else:
                self._vertices = vertices

    def bounds(self) -> typing.Tuple[float, float, float, float]:
        if self._bounds is None:
            min_x = min([p[0] for p in self.vertices])
            min_y = min([p[1] for p in self.vertices])
            max_x = max([p[0] for p in self.vertices])
            max_y = max([p[1] for p in self.vertices])
            self._bounds = (min_x, min_y, max_x, max_y)
        return self._bounds

    def sides(self):
        return self._sides

    @property
    def vertices(self) -> typing.List[Point.type]:
        if self._vertices is None:
            self._vertices = [(p.x, p.y) for p in self.points]
        return self._vertices

    @property
    def points(self) -> typing.List[sp.Point]:
        if self._points is None:
            self._points = [sp.Point(p) for p in self.vertices]
        return self._points

    @property
    def polygon(self):
        if self._polygon is None:
            self._polygon = sp.geometry.Polygon(self.points)
        return self._polygon

    def contains(self, points):
        if isinstance(points, Polygon):
            inside = all([self.polygon.contains(point) for point in points.points])
        else:
            inside = [self.polygon.contains(sp.Point(point)) for point in points]
        return inside

    def intersects(self, other: "Polygon"):
        return self.polygon.intersects(other.polygon)

    def __getitem__(self, item) -> typing.Tuple[float, float]:
        return self.vertices[item]

    def translate_rotate(self,
                         translation: Point.type,
                         rotation: float,
                         rotation_center: Point.type = (0, 0)) -> "Polygon":
        delta_x, delta_y = translation
        rotated_polygon = spa.rotate(self.polygon,
                                     rotation,
                                     origin=rotation_center,
                                     use_radians=False)
        translated_polygon: spa.Polygon = spa.translate(rotated_polygon, delta_x, delta_y)

        return Polygon(vertices=translated_polygon.exterior.coords)

    def area(self) -> float:
        return self.polygon.area



import math
import pygame
import typing
from .model import Model
from .coordinate_converter import CoordinateConverter
import colorsys
from .event import EventDispatcher
import numpy as np

def generate_distinct_colors(n):
    colors = []
    for i in range(n):
        hue = i / n  # Evenly distribute hues in the [0, 1) interval
        saturation = 0.7  # Choose a saturation level that avoids white and grays
        lightness = 0.5  # Choose a lightness level that avoids black and white
        rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
        # Convert the RGB values from [0, 1] to [0, 255] and round them to integers
        rgb = tuple(round(c * 255) for c in rgb)
        colors.append(rgb)

    return colors


class View(EventDispatcher):
    def __init__(self,
                 model: Model,
                 screen_width: int = 800,
                 flip_y: bool = True):
        pygame.init()
        self.model = model
        self.hexa_ratio = (math.sqrt(3) / 2)
        self.coordinate_converter = CoordinateConverter(screen_size=(screen_width, int(self.hexa_ratio * screen_width)),
                                                        flip_y=flip_y)
        self.render_steps: typing.List[typing.Callable[[pygame.Surface, CoordinateConverter], None]] = []
        self.render_steps_z_index: typing.List[int] = []
        self.render_steps_sequence: typing.List[int] = []
        self.screen = pygame.display.set_mode(self.coordinate_converter.screen_size)
        self.background_color = (0, 0, 0)
        self.clock = pygame.time.Clock()
        EventDispatcher.__init__(self, ["mouse_button_down",
                                        "mouse_button_up",
                                        "mouse_move",
                                        "mouse_wheel",
                                        "key_down",
                                        "key_up",
                                        "quit",
                                        "frame"])
        self.pressed_keys = pygame.key.get_pressed()
        self.draw = self.render
        self.visibility_location = (.5, .5)

    def add_render_step(self,
                        render_step: typing.Callable[[pygame.Surface, CoordinateConverter], None],
                        z_index: int = -1):
        self.render_steps.append(render_step)
        self.render_steps_z_index.append(z_index)
        self.render_steps_sequence = [i for i, z in
                                      sorted([(i, math.inf if z == -1 else z)
                                             for i, z
                                             in enumerate(self.render_steps_z_index)],
                                             key=lambda item: item[1])]

    def render(self):
        self.screen.fill(self.background_color)
        for render_step_number in self.render_steps_sequence:
            render_step = self.render_steps[render_step_number]
            render_step(self.screen, self.coordinate_converter)
        self.__process_events__()
        self.__dispatch__("frame", self.screen, self.coordinate_converter)
        pygame.display.flip()

    def __process_events__(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__dispatch__("quit")
            if event.type == pygame.MOUSEBUTTONDOWN:
                canonical_x_y = self.coordinate_converter.to_canonical(event.pos)
                self.__dispatch__("mouse_button_down", canonical_x_y)
            elif event.type == pygame.MOUSEBUTTONUP:
                canonical_x_y = self.coordinate_converter.to_canonical(event.pos)
                self.__dispatch__("mouse_button_up", canonical_x_y)
            elif event.type == pygame.MOUSEMOTION:
                canonical_x_y = self.coordinate_converter.to_canonical(event.pos)
                self.__dispatch__("mouse_move", canonical_x_y)
            elif event.type == pygame.MOUSEWHEEL:
                canonical_x_y = self.coordinate_converter.to_canonical((event.x, event.y))
                self.__dispatch__("mouse_wheel", canonical_x_y)
            elif event.type == pygame.KEYDOWN:
                self.__dispatch__("key_down", event.key)
            elif event.type == pygame.KEYUP:
                self.__dispatch__("key_up", event.key)
        self.pressed_keys = pygame.key.get_pressed()

    def get_screen(self, normalized: bool = False) -> np.ndarray:
        screen_array = pygame.surfarray.array3d(self.screen)
        screen_array = np.transpose(screen_array, (1, 0, 2))
        if normalized:
            screen_array = screen_array / 255.0
        return screen_array



import pulsekit
import pygame
from .util import Point, Direction
from .agent import Agent
from .coordinate_converter import CoordinateConverter
from .navigation import Navigation
import typing


class NavigationAgent(Agent):

    def __init__(self,
                 navigation: Navigation,
                 max_forward_speed: float,
                 max_turning_speed: float,
                 threshold: float = 0.01,
                 destination_update_rate: int = 10,
                 view_field: float = 360,
                 size: float = 0.5,
                 sprite_scale: float = 1,
                 polygon_color: typing.Tuple[int, int, int] = (0, 80, 120)):
        self.max_forward_speed = max_forward_speed
        self.max_turning_speed = max_turning_speed
        self.threshold = threshold
        self.destination_update_rate = destination_update_rate
        self.navigation = navigation
        self.new_destination = None
        self.destination = None
        self.navigation_plan_update_wait = 0
        self.destination_wait = 0
        self.path = []
        self.render_path = False
        Agent.__init__(self,
                       view_field=view_field,
                       size=size,
                       sprite_scale=sprite_scale,
                       polygon_color=polygon_color)
        self.collision = False
        self.active_navigation = True

    def next_step(self):
        if self.path:
            return self.path[0]
        return None

    def set_destination(self, destination):
        self.new_destination = destination

    def reset(self):
        self.destination = None
        self.path = []
        Agent.reset(self)

    def step(self, delta_t: float):
        if not self.active_navigation:
            return
        with pulsekit.CodeBlock("navigation_agent.step"):
            if not self.running:
                self.stop_navigation()
                return

            if self.new_destination and self.destination_wait == 0:
                self.destination = self.new_destination
                self.new_destination = None
                self.destination_wait = self.destination_update_rate
                self.navigation_plan_update_wait = self.destination_update_rate
                self.path = self.navigation.get_path(src=self.state.location, dst=self.destination)

            if self.destination and self.navigation_plan_update_wait == 0:
                self.path = self.navigation.get_path(src=self.state.location, dst=self.destination)

            if self.destination_wait:
                self.destination_wait -= 1

            if self.navigation_plan_update_wait:
                self.navigation_plan_update_wait -= 1

            if self.next_step() is not None:
                distance_error = Point.distance(src=self.state.location,
                                                dst=self.next_step())
                if distance_error < self.threshold:
                    self.path.pop(0)

            if self.next_step():
                distance_error = Point.distance(src=self.state.location,
                                                dst=self.next_step())

                normalized_distance_error = max(distance_error/.2, 1)

                destination_direction = Direction.degrees(src=self.state.location,
                                                          dst=self.next_step())
                direction_error = Direction.difference(direction1=self.state.direction,
                                                       direction2=destination_direction)
                normalized_direction_error = Direction.error_normalization(direction_error=direction_error)

                self.dynamics.forward_speed = self.max_forward_speed * normalized_direction_error * normalized_distance_error
                self.dynamics.turn_speed = self.max_turning_speed * direction_error
            else:
                self.dynamics.forward_speed = 0
                self.dynamics.turn_speed = 0

    def render(self,
               surface: pygame.Surface,
               coordinate_converter: CoordinateConverter):
        if self.render_path:
            current_point = self.state.location
            for step in self.path:
                if step is None:
                    continue
                pygame.draw.line(surface,
                                 (255, 0, 0),
                                 coordinate_converter.from_canonical(current_point),
                                 coordinate_converter.from_canonical(step),
                                 2)
                pygame.draw.circle(surface=surface,
                                   color=(0, 0, 255),
                                   center=coordinate_converter.from_canonical(step),
                                   radius=5,
                                   width=2)
                current_point = step
        Agent.render(self=self,
                     surface=surface,
                     coordinate_converter=coordinate_converter)

    def stop_navigation(self):
        self.dynamics.forward_speed = 0
        self.dynamics.turn_speed = 0
